/*
 * Copyright 2011 BitMeister Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.bitmeister.asn1.codec.ber;

import java.io.OutputStream;
import java.math.BigInteger;
import java.text.DateFormat;
import java.util.TimeZone;

import jp.bitmeister.asn1.codec.ASN1Encoder;
import jp.bitmeister.asn1.exception.ASN1EncodingException;
import jp.bitmeister.asn1.processor.ASN1Visitor;
import jp.bitmeister.asn1.type.ASN1TagClass;
import jp.bitmeister.asn1.type.ASN1TagMode;
import jp.bitmeister.asn1.type.ASN1TagValue;
import jp.bitmeister.asn1.type.ASN1Type;
import jp.bitmeister.asn1.type.CollectionType;
import jp.bitmeister.asn1.type.ConstructiveType;
import jp.bitmeister.asn1.type.ElementSpecification;
import jp.bitmeister.asn1.type.StringType;
import jp.bitmeister.asn1.type.TimeType;
import jp.bitmeister.asn1.type.TypeSpecification;
import jp.bitmeister.asn1.type.UnknownType;
import jp.bitmeister.asn1.type.builtin.ANY;
import jp.bitmeister.asn1.type.builtin.BIT_STRING;
import jp.bitmeister.asn1.type.builtin.BOOLEAN;
import jp.bitmeister.asn1.type.builtin.CHOICE;
import jp.bitmeister.asn1.type.builtin.ENUMERATED;
import jp.bitmeister.asn1.type.builtin.INTEGER;
import jp.bitmeister.asn1.type.builtin.NULL;
import jp.bitmeister.asn1.type.builtin.OBJECT_IDENTIFIER;
import jp.bitmeister.asn1.type.builtin.OCTET_STRING;
import jp.bitmeister.asn1.type.builtin.REAL;
import jp.bitmeister.asn1.type.builtin.SEQUENCE;
import jp.bitmeister.asn1.type.builtin.SEQUENCE_OF;
import jp.bitmeister.asn1.type.builtin.SET;
import jp.bitmeister.asn1.type.builtin.SET_OF;

/**
 * DER (Distinguished Encoding Rules) encoder.
 * 
 * <p>
 * {@code DerEncoder} is an implementation of {@code ASN1Encoder}. It encodes an
 * ASN.1 data to an array of {@code byte} with Distinguished Encoding Rules(DER)
 * and writes the result to {@code OutputStream} that is specified when the
 * encoder was instantiated. The result octets can be decoded by any BER decoder
 * because DER is a sub-set of Basic Encoding Rules(BER) and compatible with it.
 * </p>
 * 
 * @author WATANABE, Jun. <jwat at bitmeister.jp>
 * 
 * @see ASN1Encoder
 * @see BerDecoder
 */
public class DerEncoder implements ASN1Encoder,
		ASN1Visitor<ASN1EncodingException> {

	private OutputStream out;

	private BerOctets octets;

	/**
	 * Instantiates a DER encoder.
	 * 
	 * @param out
	 *            The {@code OutputStream} that encoded octets are written.
	 */
	public DerEncoder(OutputStream out) {
		this.out = out;
	}

	/**
	 * Encodes an ASN.1 data.
	 * 
	 * @param data
	 *            The ASN.1 data to be encoded
	 * @return The size of encoded octets.
	 * @throws ASN1EncodingException
	 *             When an error occurred while the encoding process.
	 */
	@Override
	public int encode(ASN1Type data) throws ASN1EncodingException {
		try {
			data.validate();
			encode(data, null, null);
			octets.write(this.out);
		} catch (ASN1EncodingException e) {
			throw e;
		} catch (Exception e) {
			ASN1EncodingException ex = new ASN1EncodingException();
			ex.setMessage("Exception thrown while encoding process.", e,
					data.getClass(), null, data);
			throw ex;
		}
		return octets.totalLength();
	}

	/**
	 * Encodes an ASN.1 data or an element of ASN.1 data.
	 * 
	 * @param data
	 *            The ASN.1 data to be encoded.
	 * @param tag
	 *            The tag assigned to an element to be encoded.
	 * @param typeSpec
	 *            The {@code TypeSpecification} of a type that tagged
	 *            explicitly.
	 * @throws ASN1EncodingException
	 *             When an error occurred while the encoding process.
	 */
	private void encode(ASN1Type data, ASN1TagValue tag,
			TypeSpecification typeSpec) throws ASN1EncodingException {
		if (typeSpec == null) {
			typeSpec = data.specification();
		}
		if (tag == null) {
			do {
				tag = typeSpec.tag();
				typeSpec = typeSpec.reference();
			} while (typeSpec != null && tag == null);
		}
		if (tag != null && tag.tagMode() == ASN1TagMode.EXPLICIT) {
			encode(data, null, typeSpec);
			ConstructedOctets constructed = new ConstructedOctets();
			constructed.addElement(octets);
			octets = constructed;
		} else {
			data.accept(this);
		}
		if (tag != null) {
			octets.setPrefix(encodeASN1Tag(tag.tagClass(), tag.tagNumber()),
					encodeLength(octets.octetsLength()));
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.BOOLEAN)
	 */
	@Override
	public void visit(BOOLEAN data) {
		octets = new PrimitiveOctets(data.value() ? (byte) 0xff : (byte) 0x00);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.INTEGER)
	 */
	@Override
	public void visit(INTEGER data) {
		octets = new PrimitiveOctets(data.value().toByteArray());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.ENUMERATED)
	 */
	@Override
	public void visit(ENUMERATED data) {
		visit((INTEGER) data);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.REAL)
	 */
	@Override
	public void visit(REAL data) {
		byte[] encoded;
		if (data.value() == 0) { // zero.
			encoded = new byte[0];
		} else if (data.value().isInfinite()) { // special value.
			encoded = new byte[] { data.value() == Double.POSITIVE_INFINITY ? (byte) 0x40
					: 0x41 };
		} else { // IEEE754 double form.
			long bits = Double.doubleToLongBits(data.value());

			// calculate E and M value.
			int exponent = (int) ((bits >> 52) & 0x7ff) - (1023 + 52);
			long mantissa = bits & 0xfffffffffffffL | 0x10000000000000L;
			while ((mantissa & 0x01) == 0) {
				mantissa >>= 1;
				exponent++;
			}
			int expLen = (exponent > Byte.MAX_VALUE)
					|| (exponent < Byte.MIN_VALUE) ? 2 : 1;
			byte[] mOctets = BigInteger.valueOf(mantissa).toByteArray();
			encoded = new byte[1 + expLen + mOctets.length];

			// binary encoding.
			encoded[0] = (byte) 0x80;
			if ((bits & 0x8000000000000000L) != 0) {
				// value is negative.
				encoded[0] |= 0x40;
			}
			if (expLen == 2) {
				encoded[0] |= 0x01;
				encoded[1] = (byte) (exponent >> 8);
				encoded[2] = (byte) exponent;
			} else {
				encoded[1] = (byte) exponent;
			}
			System.arraycopy(mOctets, 0, encoded, 1 + expLen, mOctets.length);
		}
		octets = new PrimitiveOctets(encoded);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.BIT_STRING)
	 */
	@Override
	public void visit(BIT_STRING data) {
		if (data.value().length == 0) {
			octets = new PrimitiveOctets((byte) 0x00);
			return;
		}
		int mod = data.value().length % 8;
		byte[] encoded = new byte[1 + data.value().length / 8
				+ (mod == 0 ? 0 : 1)];
		encoded[0] = (byte) (mod == 0 ? 0 : 8 - mod);
		int index = 1;
		int mask = 0x80;
		for (boolean b : data.value()) {
			if (b) {
				encoded[index] |= mask;
			}
			mask >>>= 1;
			if (mask == 0) {
				mask = 0x80;
				index++;
			}
		}
		octets = new PrimitiveOctets(encoded);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.OCTET_STRING)
	 */
	@Override
	public void visit(OCTET_STRING data) throws ASN1EncodingException {
		octets = new PrimitiveOctets(data.value());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.NULL)
	 */
	@Override
	public void visit(NULL data) throws ASN1EncodingException {
		octets = new PrimitiveOctets();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.SEQUENCE)
	 */
	@Override
	public void visit(SEQUENCE data) throws ASN1EncodingException {
		processConstructive(data);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.SEQUENCE_OF)
	 */
	@Override
	public void visit(SEQUENCE_OF<? extends ASN1Type> data)
			throws ASN1EncodingException {
		processCollection(data);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.SET)
	 */
	@Override
	public void visit(SET data) throws ASN1EncodingException {
		processConstructive(data);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.SET_OF)
	 */
	@Override
	public void visit(SET_OF<? extends ASN1Type> data)
			throws ASN1EncodingException {
		processCollection(data);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.CHOICE)
	 */
	@Override
	public void visit(CHOICE data) throws ASN1EncodingException {
		encode(data.selectedValue(), data.selectedTag(), null);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.OBJECT_IDENTIFIER)
	 */
	@Override
	public void visit(OBJECT_IDENTIFIER data) throws ASN1EncodingException {
		if (data.value().size() < 2 || data.value().get(0) < 0
				|| 2 < data.value().get(0) || data.value().get(1) < 0
				|| 39 < data.value().get(1)) {
			ASN1EncodingException ex = new ASN1EncodingException();
			ex.setMessage("Invalid OBJECT IDENTIFIER value.", null,
					data.getClass(), null, data);
			throw ex;
		}
		int size = 1;
		for (int i = 2; i < data.value().size(); i++) {
			if (data.value().get(i) < 0) {
				ASN1EncodingException ex = new ASN1EncodingException();
				ex.setMessage(
						"OBJECT IDENTIFIER value must be a positive number.",
						null, data.getClass(), null, data);
				throw ex;
			}
			size += sizeBy7bits(data.value().get(i));
		}
		byte[] encoded = new byte[size];
		encoded[0] = (byte) (data.value().get(0) * 40 + data.value().get(1));
		int offset = 1;
		for (int i = 2; i < data.value().size(); i++) {
			offset += encodeToMutipleOctets(encoded, offset, data.value()
					.get(i));
		}
		octets = new PrimitiveOctets(encoded);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .StringType)
	 */
	@Override
	public void visit(StringType data) throws ASN1EncodingException {
		octets = new PrimitiveOctets(data.value());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .TimeType)
	 */
	@Override
	public void visit(TimeType data) throws ASN1EncodingException {
		DateFormat form = data.form();
		String value = data.stringValue();
		form.setTimeZone(TimeZone.getTimeZone("GMT"));
		StringBuffer buf = new StringBuffer(form.format(data.date()).split(
				"[+-]")[0]);
		buf.setLength(value.split("[Z+-]")[0].length());
		if (buf.indexOf(".") > 0) {
			while (buf.lastIndexOf("0") == buf.length() - 1) {
				buf.setLength(buf.length() - 1);
			}
			if (buf.lastIndexOf(".") == buf.length() - 1) {
				buf.setLength(buf.length() - 1);
			}
		}
		buf.append('Z');
		octets = new PrimitiveOctets(buf.toString().getBytes());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .builtin.ANY)
	 */
	@Override
	public void visit(ANY data) throws ASN1EncodingException {
		encode(data.value(), null, null);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * jp.bitmeister.asn1.processor.ASN1Visitor#visit(jp.bitmeister.asn1.type
	 * .UnknownType)
	 */
	@Override
	public void visit(UnknownType data) throws ASN1EncodingException {
		octets = new PrimitiveOctets(data.value());
		octets.setPrefix(encodeASN1Tag(data.tagClass(), data.tagNumber()),
				encodeLength(data.value().length));
	}

	/**
	 * Encodes the ASN.1 tag to octets.
	 * 
	 * @param tagClass
	 *            The tag class.
	 * @param tagNumber
	 *            The tag number.
	 * @return Encoded octets.
	 */
	private byte[] encodeASN1Tag(ASN1TagClass tagClass, int tagNumber) {
		byte leading = 0;
		switch (tagClass) {
		case UNIVERSAL:
			// 0x00;
			break;
		case APPLICATION:
			leading = 0x40;
			break;
		case CONTEXT_SPECIFIC:
			leading = (byte) 0x80;
			break;
		case PRIVATE:
			leading = (byte) 0xc0;
			break;
		}
		if (octets.isConstructed()) {
			// constructed.
			leading |= 0x20;
		}
		byte[] encoded;
		if (tagNumber <= 0x1e) {
			leading |= tagNumber;
			encoded = new byte[] { leading };
		} else {
			leading |= 0x1f;
			int size = sizeBy7bits(tagNumber);
			encoded = new byte[size + 1];
			encoded[0] = leading;
			encodeToMutipleOctets(encoded, 1, tagNumber);
		}
		return encoded;
	}

	/**
	 * Encodes the length to octets.
	 * 
	 * @param length
	 *            The length.
	 * @return Encoded octets.
	 */
	private byte[] encodeLength(long length) {
		byte[] encoded;
		if (length <= 0x7f) {
			// length in short form.
			encoded = new byte[] { (byte) length };
		} else {
			// length in long form.
			byte[] value = BigInteger.valueOf(length).toByteArray();
			encoded = new byte[value.length + 1];
			encoded[0] = (byte) (value.length | 0x80);
			System.arraycopy(value, 0, encoded, 1, value.length);
		}
		return encoded;
	}

	/**
	 * Encodes each element of the {@code ConstructiveType} data.
	 * 
	 * @param data
	 *            The data to be encoded.
	 * @throws ASN1EncodingException
	 *             When an error occurred while the encoding process.
	 */
	private void processConstructive(ConstructiveType data)
			throws ASN1EncodingException {
		ConstructedOctets octets = new ConstructedOctets();
		for (ElementSpecification e : data.getElementTypeList()) {
			ASN1Type element = data.getComponent(e);
			if (element == null || !element.hasValue()) {
				continue;
			}
			encode(element, e.tag(), null);
			octets.addElement(this.octets);
		}
		this.octets = octets;
	}

	/**
	 * Encodes each component of the {@code CollectionType} data.
	 * 
	 * @param data
	 *            The data to be encoded.
	 * @throws ASN1EncodingException
	 *             When an error occurred while the encoding process.
	 */
	private void processCollection(CollectionType<?> data)
			throws ASN1EncodingException {
		ConstructedOctets octets = new ConstructedOctets();
		for (ASN1Type e : data.collection()) {
			encode(e, null, null);
			octets.addElement(this.octets);
		}
		this.octets = octets;
	}

	/**
	 * Encodes the value to long form octets.
	 * 
	 * @param dest
	 *            The array of byte where the result to be written.
	 * @param offset
	 *            The index in the {@code dest} array that indicates where the
	 *            result to be written.
	 * @param value
	 *            The integer value.
	 * @return The size of the result.
	 */
	private int encodeToMutipleOctets(byte[] dest, int offset, long value) {
		int size = sizeBy7bits(value);
		int lastIndex = offset + size - 1;
		for (int index = lastIndex; index >= offset; index--) {
			dest[index] |= (byte) (value & 0x7f);
			if (index != lastIndex) {
				dest[index] |= 0x80;
			}
			value >>= 7;
		}
		return size;
	}

	/**
	 * Estimates the size of the result that the value is encoded to long form.
	 * 
	 * @param value
	 *            The value to be encoded.
	 * @return Estimated size.
	 */
	private int sizeBy7bits(long value) {
		int size = 1;
		while ((value >>= 7) > 0) {
			size++;
		}
		return size;
	}

}
